@@ -599,60 +599,51 @@ rb_get_ssl_certfp(rb_fde_t *F, uint8_t certfp[RB_SSL_CERTFP_LEN], int method)
 {
 	gnutls_x509_crt_t cert;
 	gnutls_digest_algorithm_t algo;
-	unsigned int cert_list_size;
+	unsigned int cert_list_size = 0;
 	const gnutls_datum_t *cert_list;
-	uint8_t digest[RB_SSL_CERTFP_LEN * 2];
 	size_t digest_size;
-	int len;
-
-	if (gnutls_certificate_type_get(SSL_P(F)) != GNUTLS_CRT_X509)
-		return 0;
-
-	if (gnutls_x509_crt_init(&cert) < 0)
-		return 0;
-
-	cert_list_size = 0;
-	cert_list = gnutls_certificate_get_peers(SSL_P(F), &cert_list_size);
-	if (cert_list == NULL)
-	{
-		gnutls_x509_crt_deinit(cert);
-		return 0;
-	}
-
-	if (gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER) < 0)
-	{
-		gnutls_x509_crt_deinit(cert);
-		return 0;
-	}
 
 	switch(method)
 	{
 	case RB_SSL_CERTFP_METH_SHA1:
 		algo = GNUTLS_DIG_SHA1;
-		len = RB_SSL_CERTFP_LEN_SHA1;
 		break;
 	case RB_SSL_CERTFP_METH_SHA256:
 		algo = GNUTLS_DIG_SHA256;
-		len = RB_SSL_CERTFP_LEN_SHA256;
 		break;
 	case RB_SSL_CERTFP_METH_SHA512:
 		algo = GNUTLS_DIG_SHA512;
-		len = RB_SSL_CERTFP_LEN_SHA512;
 		break;
 	default:
 		return 0;
 	}
 
-	if (gnutls_x509_crt_get_fingerprint(cert, algo, digest, &digest_size) < 0)
+	if (gnutls_certificate_type_get(SSL_P(F)) != GNUTLS_CRT_X509)
+		return 0;
+
+	cert_list = gnutls_certificate_get_peers(SSL_P(F), &cert_list_size);
+
+	if (cert_list_size <= 0)
+		return 0;
+
+	if (gnutls_x509_crt_init(&cert) != GNUTLS_E_SUCCESS)
+		return 0;
+
+	if (gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER) != GNUTLS_E_SUCCESS)
 	{
 		gnutls_x509_crt_deinit(cert);
 		return 0;
 	}
 
-	memcpy(certfp, digest, len);
+	if (gnutls_x509_crt_get_fingerprint(cert, algo, certfp, &digest_size) != 0)
+	{
+		gnutls_x509_crt_deinit(cert);
+		return 0;
+	}
 
 	gnutls_x509_crt_deinit(cert);
-	return len;
+
+	return (int) digest_size;
 }
 
 int
@@ -673,11 +664,26 @@ rb_ssl_get_cipher(rb_fde_t *F)
 {
 	static char buf[1024];
 
-	rb_snprintf(buf, sizeof(buf), "%s-%s-%s-%s",
-		gnutls_protocol_get_name(gnutls_protocol_get_version(SSL_P(F))),
-		gnutls_kx_get_name(gnutls_kx_get(SSL_P(F))),
-		gnutls_cipher_get_name(gnutls_cipher_get(SSL_P(F))),
-		gnutls_mac_get_name(gnutls_mac_get(SSL_P(F))));
+	const char* proto_name =
+	    gnutls_protocol_get_name(gnutls_protocol_get_version(SSL_P(F)));
+
+	const char* kex_alg_name =
+	    gnutls_kx_get_name(gnutls_kx_get(SSL_P(F)));
+
+	const char* cipher_alg_name =
+	    gnutls_cipher_get_name(gnutls_cipher_get(SSL_P(F)));
+
+	const char* mac_alg_name =
+	    gnutls_mac_get_name(gnutls_mac_get(SSL_P(F)));
+
+	(void) rb_snprintf(buf, sizeof buf, "%s%s%s%s%s%s%s",
+	                   proto_name ? proto_name : "",
+	                   proto_name ? ", " : "",
+	                   kex_alg_name ? kex_alg_name : "",
+	                   kex_alg_name ? "-" : "",
+	                   cipher_alg_name ? cipher_alg_name : "",
+	                   cipher_alg_name ? "-" : "",
+	                   mac_alg_name ? mac_alg_name : "");
 
 	return buf;
 }
