@@ -61,6 +61,37 @@
 #  endif
 #endif
 
+/*
+ * More LibreSSL compatibility mess
+ * Used in rb_get_ssl_info() below.
+ */
+#if !defined(LIBRESSL_VERSION_NUMBER) && (OPENSSL_VERSION_NUMBER >= 0x10100000L)
+   /* OpenSSL 1.1.0+ */
+#  define LRB_SSL_VTEXT_COMPILETIME     OPENSSL_VERSION_TEXT
+#  define LRB_SSL_VTEXT_RUNTIME         OpenSSL_version(OPENSSL_VERSION)
+#  define LRB_SSL_VNUM_COMPILETIME      OPENSSL_VERSION_NUMBER
+#  define LRB_SSL_VNUM_RUNTIME          OpenSSL_version_num()
+#  define LRB_SSL_FULL_VERSION_INFO     1
+#else
+/*
+ * "Full version info" above means we have access to all 4 pieces of information.
+ *
+ * For the below, this is not the case; LibreSSL version number at runtime returns
+ * the wrong version number, and OpenSSL version text at compile time does not exist.
+ * Thus, we only reliably have version text at runtime, and version number at compile
+ * time.
+ */
+#  if defined(LIBRESSL_VERSION_NUMBER) && (LIBRESSL_VERSION_NUMBER >= 0x20200000L)
+     /* LibreSSL 2.2.0+ */
+#    define LRB_SSL_VTEXT_RUNTIME       SSLeay_version(SSLEAY_VERSION)
+#    define LRB_SSL_VNUM_COMPILETIME    LIBRESSL_VERSION_NUMBER
+#  else
+     /* OpenSSL < 1.1.0 or LibreSSL < 2.2.0 */
+#    define LRB_SSL_VTEXT_RUNTIME       SSLeay_version(SSLEAY_VERSION)
+#    define LRB_SSL_VNUM_COMPILETIME    SSLEAY_VERSION_NUMBER
+#  endif
+#endif
+
 static SSL_CTX *ssl_server_ctx = NULL;
 static SSL_CTX *ssl_client_ctx = NULL;
 static int libratbox_index = -1;
@@ -327,13 +358,13 @@ rb_init_ssl(void)
 {
 	char libratbox_data[] = "libratbox data";
 
-#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
 	/*
 	 * OpenSSL 1.1.0 and above automatically initialises itself with sane defaults
 	 */
+	#if defined(LIBRESSL_VERSION_NUMBER) || (OPENSSL_VERSION_NUMBER < 0x10100000L)
 	SSL_library_init();
 	SSL_load_error_strings();
-#endif
+	#endif
 
 	libratbox_index = SSL_get_ex_new_index(0, libratbox_data, NULL, NULL, NULL);
 
@@ -345,6 +376,9 @@ rb_setup_ssl_server(const char *cert, const char *keyfile, const char *dhfile, c
 {
 	const char libratbox_ciphers[] = "kEECDH+HIGH:kEDH+HIGH:HIGH:!aNULL";
 
+	SSL_CTX *ssl_server_ctx_new;
+	SSL_CTX *ssl_client_ctx_new;
+
 	#ifdef LRB_HAVE_TLS_SET_CURVES
 	const char libratbox_curves[] = "P-521:P-384:P-256";
 	#endif
@@ -364,67 +398,63 @@ rb_setup_ssl_server(const char *cert, const char *keyfile, const char *dhfile, c
 	if(cipher_list == NULL)
 		cipher_list = libratbox_ciphers;
 
-	if(ssl_server_ctx)
-		SSL_CTX_free(ssl_server_ctx);
-
-	if(ssl_client_ctx)
-		SSL_CTX_free(ssl_client_ctx);
-
 	#ifdef LRB_HAVE_TLS_METHOD_API
-	ssl_server_ctx = SSL_CTX_new(TLS_server_method());
-	ssl_client_ctx = SSL_CTX_new(TLS_client_method());
+	if((ssl_server_ctx_new = SSL_CTX_new(TLS_server_method())) == NULL)
 	#else
-	ssl_server_ctx = SSL_CTX_new(SSLv23_server_method());
-	ssl_client_ctx = SSL_CTX_new(SSLv23_client_method());
+	if((ssl_server_ctx_new = SSL_CTX_new(SSLv23_server_method())) == NULL)
 	#endif
-
-	if(ssl_server_ctx == NULL)
 	{
 		rb_lib_log("rb_init_openssl: Unable to initialize OpenSSL server context: %s",
 			   get_ssl_error(ERR_get_error()));
 		return 0;
 	}
 
-	if(ssl_client_ctx == NULL)
+	#ifdef LRB_HAVE_TLS_METHOD_API
+	if((ssl_client_ctx_new = SSL_CTX_new(TLS_client_method())) == NULL)
+	#else
+	if((ssl_client_ctx_new = SSL_CTX_new(SSLv23_client_method())) == NULL)
+	#endif
 	{
 		rb_lib_log("rb_init_openssl: Unable to initialize OpenSSL client context: %s",
 			   get_ssl_error(ERR_get_error()));
+
+		SSL_CTX_free(ssl_server_ctx_new);
 		return 0;
 	}
 
 	#ifndef LRB_HAVE_TLS_METHOD_API
-	SSL_CTX_set_options(ssl_server_ctx, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
-	SSL_CTX_set_options(ssl_client_ctx, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
+	SSL_CTX_set_options(ssl_server_ctx_new, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
+	SSL_CTX_set_options(ssl_client_ctx_new, SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
 	#endif
 
 	#ifdef SSL_OP_SINGLE_DH_USE
-	SSL_CTX_set_options(ssl_server_ctx, SSL_OP_SINGLE_DH_USE);
+	SSL_CTX_set_options(ssl_server_ctx_new, SSL_OP_SINGLE_DH_USE);
 	#endif
 
 	#ifdef SSL_OP_SINGLE_ECDH_USE
-	SSL_CTX_set_options(ssl_server_ctx, SSL_OP_SINGLE_ECDH_USE);
+	SSL_CTX_set_options(ssl_server_ctx_new, SSL_OP_SINGLE_ECDH_USE);
 	#endif
 
 	#ifdef SSL_OP_NO_TICKET
-	SSL_CTX_set_options(ssl_server_ctx, SSL_OP_NO_TICKET);
-	SSL_CTX_set_options(ssl_client_ctx, SSL_OP_NO_TICKET);
+	SSL_CTX_set_options(ssl_server_ctx_new, SSL_OP_NO_TICKET);
+	SSL_CTX_set_options(ssl_client_ctx_new, SSL_OP_NO_TICKET);
 	#endif
 
 	#ifdef SSL_OP_CIPHER_SERVER_PREFERENCE
-	SSL_CTX_set_options(ssl_server_ctx, SSL_OP_CIPHER_SERVER_PREFERENCE);
+	SSL_CTX_set_options(ssl_server_ctx_new, SSL_OP_CIPHER_SERVER_PREFERENCE);
 	#endif
 
 	#ifdef LRB_HAVE_TLS_ECDH_AUTO
-	SSL_CTX_set_ecdh_auto(ssl_server_ctx, 1);
+	SSL_CTX_set_ecdh_auto(ssl_server_ctx_new, 1);
 	#endif
 
 	#ifdef LRB_HAVE_TLS_SET_CURVES
-	SSL_CTX_set1_curves_list(ssl_server_ctx, libratbox_curves);
-	SSL_CTX_set1_curves_list(ssl_client_ctx, libratbox_curves);
+	SSL_CTX_set1_curves_list(ssl_server_ctx_new, libratbox_curves);
+	SSL_CTX_set1_curves_list(ssl_client_ctx_new, libratbox_curves);
 	#endif
 
-	SSL_CTX_set_verify(ssl_server_ctx, SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE, verify_accept_all_cb);
-	SSL_CTX_set_session_cache_mode(ssl_server_ctx, SSL_SESS_CACHE_OFF);
+	SSL_CTX_set_verify(ssl_server_ctx_new, SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE, verify_accept_all_cb);
+	SSL_CTX_set_session_cache_mode(ssl_server_ctx_new, SSL_SESS_CACHE_OFF);
 
 	/*
 	 * Set manual ECDHE curve on OpenSSL 1.0.0 & 1.0.1, but make sure it's actually available
@@ -432,54 +462,68 @@ rb_setup_ssl_server(const char *cert, const char *keyfile, const char *dhfile, c
 	#if (OPENSSL_VERSION_NUMBER >= 0x10000000L) && (OPENSSL_VERSION_NUMBER < 0x10002000L) && !defined(OPENSSL_NO_ECDH)
 	EC_KEY *key = EC_KEY_new_by_curve_name(NID_secp384r1);
 	if(key) {
-		SSL_CTX_set_tmp_ecdh(ssl_server_ctx, key);
+		SSL_CTX_set_tmp_ecdh(ssl_server_ctx_new, key);
 		EC_KEY_free(key);
 	}
 	#endif
 
-	SSL_CTX_set_cipher_list(ssl_server_ctx, cipher_list);
-	SSL_CTX_set_cipher_list(ssl_client_ctx, cipher_list);
+	SSL_CTX_set_cipher_list(ssl_server_ctx_new, cipher_list);
+	SSL_CTX_set_cipher_list(ssl_client_ctx_new, cipher_list);
 
-	if(!SSL_CTX_use_certificate_chain_file(ssl_server_ctx, cert) || !SSL_CTX_use_certificate_chain_file(ssl_client_ctx, cert))
+	if(!SSL_CTX_use_certificate_chain_file(ssl_server_ctx_new, cert) || !SSL_CTX_use_certificate_chain_file(ssl_client_ctx_new, cert))
 	{
 		rb_lib_log("rb_setup_ssl_server: Error loading certificate file [%s]: %s", cert,
 			   get_ssl_error(ERR_get_error()));
+
+		SSL_CTX_free(ssl_server_ctx_new);
+		SSL_CTX_free(ssl_client_ctx_new);
 		return 0;
 	}
 
-	if(!SSL_CTX_use_PrivateKey_file(ssl_server_ctx, keyfile, SSL_FILETYPE_PEM) || !SSL_CTX_use_PrivateKey_file(ssl_client_ctx, keyfile, SSL_FILETYPE_PEM))
+	if(!SSL_CTX_use_PrivateKey_file(ssl_server_ctx_new, keyfile, SSL_FILETYPE_PEM) || !SSL_CTX_use_PrivateKey_file(ssl_client_ctx_new, keyfile, SSL_FILETYPE_PEM))
 	{
 		rb_lib_log("rb_setup_ssl_server: Error loading keyfile [%s]: %s", keyfile,
 			   get_ssl_error(ERR_get_error()));
+
+		SSL_CTX_free(ssl_server_ctx_new);
+		SSL_CTX_free(ssl_client_ctx_new);
 		return 0;
 	}
 
 	if(dhfile != NULL)
 	{
 		/* DH parameters aren't necessary, but they are nice..if they didn't pass one..that is their problem */
-		BIO *bio = BIO_new_file(dhfile, "r");
-		if(bio != NULL)
+		FILE *fp = fopen(dhfile, "r");
+		DH *dh = NULL;
+
+		if(fp == NULL)
 		{
-			DH *dh = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
-			if(dh == NULL)
-			{
-				rb_lib_log
-					("rb_setup_ssl_server: Error loading DH params file [%s]: %s",
-					 dhfile, get_ssl_error(ERR_get_error()));
-				BIO_free(bio);
-				return 0;
-			}
-			BIO_free(bio);
-			SSL_CTX_set_tmp_dh(ssl_server_ctx, dh);
-			DH_free(dh);
+			rb_lib_log("rb_setup_ssl_server: Error loading DH params file [%s]: %s",
+			           dhfile, strerror(errno));
+		}
+		else if(PEM_read_DHparams(fp, &dh, NULL, NULL) == NULL)
+		{
+			rb_lib_log("rb_setup_ssl_server: Error loading DH params file [%s]: %s",
+			           dhfile, get_ssl_error(ERR_get_error()));
+			fclose(fp);
 		}
 		else
 		{
-			rb_lib_log("rb_setup_ssl_server: Error loading DH params file [%s]: %s",
-				   dhfile, get_ssl_error(ERR_get_error()));
+			SSL_CTX_set_tmp_dh(ssl_server_ctx_new, dh);
+			DH_free(dh);
+			fclose(fp);
 		}
 	}
 
+	if(ssl_server_ctx)
+		SSL_CTX_free(ssl_server_ctx);
+
+	if(ssl_client_ctx)
+		SSL_CTX_free(ssl_client_ctx);
+
+	ssl_server_ctx = ssl_server_ctx_new;
+	ssl_client_ctx = ssl_client_ctx_new;
+
 	return 1;
 }
 
@@ -768,37 +812,34 @@ rb_supports_ssl(void)
 void
 rb_get_ssl_info(char *buf, size_t len)
 {
-#if (OPENSSL_VERSION_NUMBER >= 0x10100000L)
-	if (OpenSSL_version_num() == OPENSSL_VERSION_NUMBER)
-		rb_snprintf(buf, len, "OpenSSL: 0x%lx, %s",
-		            OPENSSL_VERSION_NUMBER, OPENSSL_VERSION_TEXT);
+#ifdef LRB_SSL_FULL_VERSION_INFO
+	if (LRB_SSL_VNUM_RUNTIME == LRB_SSL_VNUM_COMPILETIME)
+		rb_snprintf(buf, len, "OpenSSL: compiled 0x%lx, library %s",
+		            LRB_SSL_VNUM_COMPILETIME, LRB_SSL_VTEXT_COMPILETIME);
 	else
 		rb_snprintf(buf, len, "OpenSSL: compiled (0x%lx, %s), library (0x%lx, %s)",
-		            OPENSSL_VERSION_NUMBER, OPENSSL_VERSION_TEXT,
-		            OpenSSL_version_num(), OpenSSL_version(OPENSSL_VERSION));
+		            LRB_SSL_VNUM_COMPILETIME, LRB_SSL_VTEXT_COMPILETIME,
+		            LRB_SSL_VNUM_RUNTIME, LRB_SSL_VTEXT_RUNTIME);
 #else
-	if (SSLeay() == SSLEAY_VERSION_NUMBER)
-		rb_snprintf(buf, len, "OpenSSL: 0x%lx, %s",
-		            SSLeay(), SSLeay_version(SSLEAY_VERSION));
-	else
-		rb_snprintf(buf, len, "OpenSSL: compiled (0x%lx, %s), library (0x%lx, %s)",
-		            SSLEAY_VERSION_NUMBER, "???",
-		            SSLeay(), SSLeay_version(SSLEAY_VERSION));
+	rb_snprintf(buf, len, "OpenSSL: compiled 0x%lx, library %s",
+	            LRB_SSL_VNUM_COMPILETIME, LRB_SSL_VTEXT_RUNTIME);
 #endif
 }
 
 const char *
 rb_ssl_get_cipher(rb_fde_t *F)
 {
-	const SSL_CIPHER *sslciph;
-
 	if(F == NULL || F->ssl == NULL)
 		return NULL;
 
-	if((sslciph = SSL_get_current_cipher(F->ssl)) == NULL)
-		return NULL;
+	static char buf[512];
 
-	return SSL_CIPHER_get_name(sslciph);
+	const char *version = SSL_get_version(F->ssl);
+	const char *cipher = SSL_get_cipher_name(F->ssl);
+
+	(void) rb_snprintf(buf, sizeof buf, "%s, %s", version, cipher);
+
+	return buf;
 }
 
 #endif /* HAVE_OPESSL */
