@@ -255,7 +255,7 @@ try_connections(void *unused)
 			continue;
 
 		/* don't allow ssl connections if ssl isn't setup */
-		if(ServerConfSSL(tmp_p) && (!ssl_ok || !get_ssld_count()))
+		if(ServerConfSSL(tmp_p) && (!ircd_ssl_ok || !get_ssld_count()))
 			continue;
 
 		cltmp = tmp_p->class;
@@ -800,7 +800,7 @@ server_estab(struct Client *client_p)
 			   EmptyString(server_p->spasswd) ? "*" : server_p->spasswd, TS_CURRENT, me.id);
 
 		/* pass info to new server */
-		send_capabilities(client_p, default_server_capabs
+		send_capabilities(client_p, default_server_capabs | CAP_MASK
 				  | (ServerConfCompressed(server_p) ? CAP_ZIP_SUPPORTED : 0)
 				  | (ServerConfTb(server_p) ? CAP_TB : 0));
 
@@ -1357,7 +1357,7 @@ serv_connect_callback(rb_fde_t *F, int status, void *data)
 		   EmptyString(server_p->spasswd) ? "*" : server_p->spasswd, TS_CURRENT, me.id);
 
 	/* pass my info to the new server */
-	send_capabilities(client_p, default_server_capabs
+	send_capabilities(client_p, default_server_capabs | CAP_MASK
 			  | (ServerConfCompressed(server_p) ? CAP_ZIP_SUPPORTED : 0)
 			  | (ServerConfTb(server_p) ? CAP_TB : 0));
 
